from numpy._core._exceptions import _UFuncNoLoopError, UFuncTypeError
from tests.conftest import *
import numpy
import pytest
from typing import Any

@pytest.mark.parametrize("description,expected,arrayTarget,comparand", [
    ("Simple array under limit", numpy.array([0.3, -0.5]), numpy.array([0.3, -0.5]), 0.8),
    ("Simple array at limit", numpy.array([1.3, -1.3]), numpy.array([1.3, -1.3]), 1.3),
    # ("Simple array over limit", numpy.array([2.1, -2.1]), numpy.array([3.4, -3.4]), 2.1),
    ("Array comparand under limit", numpy.array([0.5, -0.8]), numpy.array([0.5, -0.8]), numpy.array([1.3, 2.1])),
    ("Array comparand mixed limits", numpy.array([0.3, 2.1, -3.4]), numpy.array([0.3, 5.5, -3.4]), numpy.array([0.5, 2.1, 3.4])),
    ("Zero array", numpy.zeros(3), numpy.zeros(3), 1.3),
    # ("Zero comparand", numpy.zeros(2), numpy.array([0.8, -1.3]), 0.0),
    ("2D array under limit", numpy.array([[0.3, -0.8], [1.3, -2.1]]), numpy.array([[0.3, -0.8], [1.3, -2.1]]), 3.4),
    ("2D array over limit", numpy.array([[2.1, -3.4], [5.5, -0.8]]), numpy.array([[2.1, -3.4], [5.5, -0.8]]), 5.5),
    ("Non-array input", TypeError, 0.5, 0.8),
    ("Mismatched shapes", IndexError, numpy.array([1.3, 2.1]), numpy.array([[3.4]])),
    ("Invalid dtype", _UFuncNoLoopError, numpy.array(['N', 'E']), 1.3),
    # ("Invalid dtype", type[UFuncTypeError], numpy.array(['S', 'W']), 2.1),
], ids=lambda x: x if isinstance(x, str) else "")
def testApplyHardLimit(description: str, expected: NDArray[Any] | type[Exception], arrayTarget: NDArray[Any] | float, comparand: float | NDArray[Any]) -> None:
    """Test applyHardLimit with various inputs."""
    prototype_numpyAllClose(expected, None, None, applyHardLimit, arrayTarget, comparand)

@pytest.mark.parametrize("description,expected,arrayTarget,comparand,penalty", [
    ("Simple complex under limit", numpy.array([0.3+0.5j, -0.8-1.3j]), numpy.array([0.3+0.5j, -0.8-1.3j]), numpy.array([2.1, 3.4]), 0.5),
    ("Simple complex at limit", numpy.array([0.8+1.3j, -2.1-3.4j]), numpy.array([0.8+1.3j, -2.1-3.4j]), numpy.array([3.4, 5.5]), 0.8),
    # ("Simple complex over limit", numpy.array([(1.3+2.1j)*numpy.sqrt(2), (-3.4-5.5j)*numpy.sqrt(2)]), numpy.array([2.1+3.4j, -5.5-0.8j]), numpy.array([0.3, 0.5]), 1.3),
    # ("Non-complex array", numpy.array([0.3, 0.8]), numpy.array([1.3, 2.1]), numpy.array([3.4, 5.5]), 2.1),
    ("Invalid penalty", TypeError, numpy.array([0.5+0.8j, 1.3+2.1j]), numpy.array([3.4, 5.5]), "invalid"),
    ("Zero complex array", numpy.zeros(3, dtype=complex), numpy.zeros(3, dtype=complex), numpy.array([0.3, 0.5, 0.8])*1.3, 3.4),
    ("Mismatched shapes", IndexError, numpy.array([0.3+0.5j, 0.8+1.3j]), numpy.array([[2.1]]), 5.5),
    # ("2D complex array", numpy.array([[0.3+0.5j, (0.8+1.3j)*numpy.sqrt(2)], [-2.1-3.4j, (-5.5-0.8j)*numpy.sqrt(2)]], dtype=numpy.complex128),
    #                     numpy.array([[0.3+0.5j, 1.3+2.1j], [-3.4-5.5j, -0.8-0.3j]], dtype=numpy.complex128),
    #                     numpy.array([[0.5, 0.8], [1.3, 2.1]], dtype=numpy.float64), 3.4),
    # ("Over limit with penalty=2", numpy.array([(0.3+0.5j)*2/numpy.sqrt(8), (-0.8-1.3j)*2/numpy.sqrt(8)]),
    #                             numpy.array([2.1+3.4j, -5.5-0.8j]), numpy.array([0.3, 0.5]), 2.0),
    # ("Complex comparand", numpy.array([(1.3+2.1j)*numpy.sqrt(2), (-3.4-5.5j)*numpy.sqrt(2)]),
    #                     numpy.array([0.3+0.5j, -0.8-1.3j]), numpy.array([2.1+3.4j, 5.5+0.8j]), 0.5),
], ids=lambda x: x if isinstance(x, str) else "")
def testApplyHardLimitComplexValued(description: str
, expected: Any
, arrayTarget: NDArray[Any]
, comparand: NDArray[Any] | NDArray[float64]
, penalty: float | str
) -> None:
    prototype_numpyAllClose(expected, None, None, applyHardLimitComplexValued, arrayTarget, comparand, penalty)



"""
Provides parameter and input validation, integer parsing, and concurrency handling utilities.
"""
from collections.abc import Iterable, Sized
from dataclasses import dataclass
from typing import Any
import charset_normalizer
import multiprocessing

@dataclass
class ErrorMessageContext:
	parameterValue: Any = None
	parameterValueType: str | None = None
	containerType: str | None = None
	isElement: bool = False

def _constructErrorMessage(context: ErrorMessageContext, parameterName: str, parameterType: type[Any] | None) -> str:
	"""Constructs error message from available context using template:
	I received ["value" | a value | None] [of type `type` | None] [as an element in | None] [a `containerType` type | None] but `parameterName` must have integers [in type(s) `parameterType` | None].

	Hypothetically, this is a prototype that can be generalized to other functions. In this package and a few of my other packages, I have developed standardized error messages, but those are quite different from this. I will certainly continue to develop this kind of functionality, and this function will influence things.
	"""
	messageParts = ["I received "]

	if context.parameterValue is not None and not isinstance(context.parameterValue, (bytes, bytearray, memoryview)):
		messageParts.append(f'"{context.parameterValue}"')
	else:
		messageParts.append("a value")

	if context.parameterValueType:
		messageParts.append(f" of type `{context.parameterValueType}`")

	if context.isElement:
		messageParts.append(" as an element in")

	if context.containerType:
		messageParts.append(f" a `{context.containerType}` type")

	messageParts.append(f" but {parameterName} must have integers")

	if parameterType:
		messageParts.append(f" in type(s) `{parameterType}`")

	return "".join(messageParts)

# NOTE: For the `limit` parameter, the most precise type is `Union[bool, float, int, None]`
# because `None` is in fact a valid option for the parameter, which is why I'm not using `Optional[Union[int, float, bool]]`.
def defineConcurrencyLimit(limit: bool | float | int | None) -> int:
	"""
	Determines the concurrency limit based on the provided parameter. This package has Pytest tests you can import and run on this function. `from Z0Z_tools.pytest_parseParameters import makeTestSuiteConcurrencyLimit`

	Parameters:
		limit: Whether and how to limit CPU usage. Accepts True/False, an integer count, or a fraction of total CPUs.
			   Positive and negative values have different behaviors, see code for details.

	Returns:
		concurrencyLimit: The calculated concurrency limit, ensuring it is at least 1.

	Notes:
		If you want to be extra nice to your users, consider using `Z0Z_tools.oopsieKwargsie()` to handle
	malformed inputs. For example:

	```
	if not (CPUlimit is None or isinstance(CPUlimit, (bool, int, float))):
		CPUlimit = oopsieKwargsie(CPUlimit)
	```

	Example parameter:
		from typing import Optional, Union
		CPUlimit: Optional[Union[int, float, bool]] = None

	Example parameter:
		from typing import Union
		CPUlimit: Union[bool, float, int, None]

	Example docstring:

	Parameters:
		CPUlimit: whether and how to limit the CPU usage. See notes for details.

	Limits on CPU usage `CPUlimit`:
		- `False`, `None`, or `0`: No limits on CPU usage; uses all available CPUs. All other values will potentially limit CPU usage.
		- `True`: Yes, limit the CPU usage; limits to 1 CPU.
		- Integer `>= 1`: Limits usage to the specified number of CPUs.
		- Decimal value (`float`) between 0 and 1: Fraction of total CPUs to use.
		- Decimal value (`float`) between -1 and 0: Fraction of CPUs to *not* use.
		- Integer `<= -1`: Subtract the absolute value from total CPUs.
	"""
	cpuTotal = multiprocessing.cpu_count()
	concurrencyLimit = cpuTotal

	if isinstance(limit, str):
		limitFromString = oopsieKwargsie(limit) # type: ignore
		if isinstance(limitFromString, str):
			try:
				limit = float(limitFromString)
			except ValueError:
				raise ValueError(f"I received '{limitFromString}', but it must be a number, True, False, or None.")
		else:
			limit = limitFromString
	if isinstance(limit, float) and abs(limit) >= 1:
		limit = round(limit)
	match limit:
		case None | False | 0:
			pass
		case True:
			concurrencyLimit = 1
		case _ if limit >= 1:
			concurrencyLimit = int(limit)
		case _ if 0 < limit < 1:
			concurrencyLimit = round(limit * cpuTotal)
		case _ if -1 < limit < 0:
			concurrencyLimit = cpuTotal - abs(round(limit * cpuTotal))
		case _ if limit <= -1:
			concurrencyLimit = cpuTotal - abs(int(limit))
		case _: pass # To placate Pylance "strict" erroneously saying not all int or float are covered.

	return max(int(concurrencyLimit), 1)

def intInnit(listInt_Allegedly: Iterable[Any], parameterName: str | None = None, parameterType: type[Any] | None = None) -> list[int]:
	"""
	Validates and converts input values to a list of integers.

	Accepts various numeric types and attempts to convert them into integers while providing descriptive error messages.

	Parameters
		listInt_Allegedly: The input sequence that should contain integer-compatible values.
			Accepts integers, strings, floats, complex numbers, and binary data.
			Rejects boolean values and non-integer numeric values.

		parameterName ('the parameter'): Name of the parameter from your function for which this function is validating the input validated: if there is an error message, it provides context to your user. Defaults to 'the parameter'.

		parameterType: Expected type(s) of the parameter, used in error messages.

	Returns
		A list containing validated integers.

	Raises
		ValueError: When the input is empty or contains non-integer compatible values.
		TypeError: When an element is a boolean or incompatible type.
		RuntimeError: If the input sequence length changes during iteration.

	Notes
		This package includes Pytest tests that can be imported and run:
		`from Z0Z_tools.pytest_parseParameters import makeTestSuiteIntInnit`

		The function performs strict validation and follows fail-early principles to catch potential issues before they become catastrophic.
	"""
	parameterName = parameterName or 'the parameter'
	parameterType = parameterType or list

	if not listInt_Allegedly:
		raise ValueError(f"I did not receive a value for {parameterName}, but it is required.")

	# Be nice: assume the input container is valid and every element is valid.
	# Nevertheless, this is a "fail-early" step, so reject ambiguity and try to induce errors now that could be catastrophic later.
	# And, be helpful: provide a meaningful error message so the user can correct the issue.
	try:
		iter(listInt_Allegedly)
		lengthInitial = None
		if isinstance(listInt_Allegedly, Sized):
			lengthInitial = len(listInt_Allegedly)

		listValidated: list[int] = []

		for allegedInt in listInt_Allegedly:
			errorMessageContext = ErrorMessageContext(
				parameterValue = allegedInt,
				parameterValueType = type(allegedInt).__name__,
				isElement = True
			)

			# Always rejected
			if isinstance(allegedInt, bool):
				raise TypeError(errorMessageContext)

			# The Python type is not `int`, but maybe the value is clearly an integer.
			# Through a series of conversions, allow data to cascade down into either an `int` or a meaningful error message.

			if isinstance(allegedInt, (bytes, bytearray, memoryview)):
				errorMessageContext.parameterValue = None  # Don't expose binary data in error messages
				if isinstance(allegedInt, memoryview):
					allegedInt = allegedInt.tobytes()
				decodedString = charset_normalizer.from_bytes(allegedInt).best()
				if not decodedString:
					raise ValueError(errorMessageContext)
				allegedInt = errorMessageContext.parameterValue = str(decodedString)

			if isinstance(allegedInt, complex):
				if allegedInt.imag != 0:
					raise ValueError(errorMessageContext)
				allegedInt = float(allegedInt.real)
			elif isinstance(allegedInt, str):
				allegedInt = float(allegedInt.strip())

			if isinstance(allegedInt, float):
				if not float(allegedInt).is_integer():
					raise ValueError(errorMessageContext)
				allegedInt = int(allegedInt)
			else:
				allegedInt = int(allegedInt)

			listValidated.append(allegedInt)

			if lengthInitial is not None and isinstance(listInt_Allegedly, Sized):
				if len(listInt_Allegedly) != lengthInitial:
					raise RuntimeError((lengthInitial, len(listInt_Allegedly)))

		return listValidated

	except (TypeError, ValueError) as ERRORmessage:
		if isinstance(ERRORmessage.args[0], ErrorMessageContext):
			context = ERRORmessage.args[0]
			if not context.containerType:
				context.containerType = type(listInt_Allegedly).__name__
			message = _constructErrorMessage(context, parameterName, parameterType)
			raise type(ERRORmessage)(message) from None
		# If it's not our Exception, don't molest it
		raise

	except RuntimeError as ERRORruntime:
		lengthInitial, lengthCurrent = ERRORruntime.args[0]
		raise RuntimeError(
			f"The input sequence {parameterName} was modified during iteration. "
			f"Initial length {lengthInitial}, current length {lengthCurrent}."
		) from None

def oopsieKwargsie(huh: Any) -> bool | None | str:
	"""
	If a calling function passes a `str` to a parameter that shouldn't receive a `str`, `oopsieKwargsie()` might help you avoid an Exception. It tries to interpret the string as `True`, `False`, or `None`. This package has Pytest tests you can import and run on this function. `from Z0Z_tools.pytest_parseParameters import makeTestSuiteOopsieKwargsie`

	Parameters:
		huh: The input string to be parsed.

	Returns:
		(bool | None | str): The reserved keywords `True`, `False`, or `None` or the original string, `huh`.
	"""
	if not isinstance(huh, str):
		try:
			huh = str(huh)
		except Exception:
			return huh
	formatted = huh.strip().title()
	if formatted == str(True):
		return True
	elif formatted == str(False):
		return False
	elif formatted == str(None):
		return None
	else:
		return huh

if __name__ == '__main__':
	# Frankly, I cannot remember the precise reason I put this in some modules. It solved the concurrency problem I was having at the time,
	# but it felt like a hack at the time and it feels even more like a hack now. I suspect I will eventually learn enough so that I can
	# come full circle: know why I added it, know how I already fixed the real issue, and know that I can safely remove this.
	multiprocessing.set_start_method('spawn')

	# Well, actually, I don't want to be programming for so long that I learn that much. I want to heal and do things in my areas of competency.



import numba.core.registry
import numba.np.ufunc.gufunc

def _resample_loop(x, t_out, interp_win, interp_delta, num_table, scale, y): ...

_resample_loop_p: numba.core.registry.CPUDispatcher
_resample_loop_s: numba.core.registry.CPUDispatcher
resample_f_p: numba.np.ufunc.gufunc.GUFunc
resample_f_s: numba.np.ufunc.gufunc.GUFunc



from collections.abc import Callable
from typing import ParamSpec, TypeVar
from tests.conftest import PytestFor_defineConcurrencyLimit, PytestFor_intInnit, PytestFor_oopsieKwargsie
import pytest

P = ParamSpec('P')
R = TypeVar('R')

@pytest.mark.parametrize("nameOfTest,callablePytest", PytestFor_defineConcurrencyLimit())
def testConcurrencyLimit(nameOfTest: str, callablePytest: Callable[P, R], *arguments: P.args, **keywordArguments: P.kwargs) -> None:
	callablePytest(*arguments, **keywordArguments)

@pytest.mark.parametrize("nameOfTest,callablePytest", PytestFor_intInnit())
def testIntInnit(nameOfTest: str, callablePytest: Callable[P, R], *arguments: P.args, **keywordArguments: P.kwargs) -> None:
	callablePytest(*arguments, **keywordArguments)

@pytest.mark.parametrize("nameOfTest,callablePytest", PytestFor_oopsieKwargsie())
def testOopsieKwargsie(nameOfTest: str, callablePytest: Callable[P, R], *arguments: P.args, **keywordArguments: P.kwargs) -> None:
	callablePytest(*arguments, **keywordArguments)



import _cffi_backend
from typing import ClassVar

__version__: str
SEEK_SET: int
SEEK_CUR: int
SEEK_END: int
_ffi: _cffi_backend.FFI
_str_types: dict
_formats: dict
_subtypes: dict
_endians: dict
_default_subtypes: dict
_ffi_types: dict
_bitrate_modes: dict
_packaged_libname: str
_path: str
_full_path: str
__libsndfile_version__: str
def read(file, frames: int = ..., start: int = ..., stop, dtype: str = ..., always_2d: bool = ..., fill_value, out, samplerate, channels, format, subtype, endian, closefd: bool = ...):
    """Provide audio data from a sound file as NumPy array.

    By default, the whole file is read from the beginning, but the
    position to start reading can be specified with *start* and the
    number of frames to read can be specified with *frames*.
    Alternatively, a range can be specified with *start* and *stop*.

    If there is less data left in the file than requested, the rest of
    the frames are filled with *fill_value*.
    If no *fill_value* is specified, a smaller array is returned.

    Parameters
    ----------
    file : str or int or file-like object
        The file to read from.  See `SoundFile` for details.
    frames : int, optional
        The number of frames to read. If *frames* is negative, the whole
        rest of the file is read.  Not allowed if *stop* is given.
    start : int, optional
        Where to start reading.  A negative value counts from the end.
    stop : int, optional
        The index after the last frame to be read.  A negative value
        counts from the end.  Not allowed if *frames* is given.
    dtype : {'float64', 'float32', 'int32', 'int16'}, optional
        Data type of the returned array, by default ``'float64'``.
        Floating point audio data is typically in the range from
        ``-1.0`` to ``1.0``.  Integer data is in the range from
        ``-2**15`` to ``2**15-1`` for ``'int16'`` and from ``-2**31`` to
        ``2**31-1`` for ``'int32'``.

        .. note:: Reading int values from a float file will *not*
            scale the data to [-1.0, 1.0). If the file contains
            ``np.array([42.6], dtype='float32')``, you will read
            ``np.array([43], dtype='int32')`` for ``dtype='int32'``.

    Returns
    -------
    audiodata : `numpy.ndarray` or type(out)
        A two-dimensional (frames x channels) NumPy array is returned.
        If the sound file has only one channel, a one-dimensional array
        is returned.  Use ``always_2d=True`` to return a two-dimensional
        array anyway.

        If *out* was specified, it is returned.  If *out* has more
        frames than available in the file (or if *frames* is smaller
        than the length of *out*) and no *fill_value* is given, then
        only a part of *out* is overwritten and a view containing all
        valid frames is returned.
    samplerate : int
        The sample rate of the audio file.

    Other Parameters
    ----------------
    always_2d : bool, optional
        By default, reading a mono sound file will return a
        one-dimensional array.  With ``always_2d=True``, audio data is
        always returned as a two-dimensional array, even if the audio
        file has only one channel.
    fill_value : float, optional
        If more frames are requested than available in the file, the
        rest of the output is be filled with *fill_value*.  If
        *fill_value* is not specified, a smaller array is returned.
    out : `numpy.ndarray` or subclass, optional
        If *out* is specified, the data is written into the given array
        instead of creating a new array.  In this case, the arguments
        *dtype* and *always_2d* are silently ignored!  If *frames* is
        not given, it is obtained from the length of *out*.
    samplerate, channels, format, subtype, endian, closefd
        See `SoundFile`.

    Examples
    --------
    >>> import soundfile as sf
    >>> data, samplerate = sf.read('stereo_file.wav')
    >>> data
    array([[ 0.71329652,  0.06294799],
           [-0.26450912, -0.38874483],
           ...
           [ 0.67398441, -0.11516333]])
    >>> samplerate
    44100

    """
def write(file, data, samplerate, subtype, endian, format, closefd: bool = ..., compression_level, bitrate_mode):
    """Write data to a sound file.

    .. note:: If *file* exists, it will be truncated and overwritten!

    Parameters
    ----------
    file : str or int or file-like object
        The file to write to.  See `SoundFile` for details.
    data : array_like
        The data to write.  Usually two-dimensional (frames x channels),
        but one-dimensional *data* can be used for mono files.
        Only the data types ``'float64'``, ``'float32'``, ``'int32'``
        and ``'int16'`` are supported.

        .. note:: The data type of *data* does **not** select the data
                  type of the written file. Audio data will be
                  converted to the given *subtype*. Writing int values
                  to a float file will *not* scale the values to
                  [-1.0, 1.0). If you write the value ``np.array([42],
                  dtype='int32')``, to a ``subtype='FLOAT'`` file, the
                  file will then contain ``np.array([42.],
                  dtype='float32')``.

    samplerate : int
        The sample rate of the audio data.
    subtype : str, optional
        See `default_subtype()` for the default value and
        `available_subtypes()` for all possible values.

    Other Parameters
    ----------------
    format, endian, closefd, compression_level, bitrate_mode
        See `SoundFile`.

    Examples
    --------
    Write 10 frames of random data to a new file:

    >>> import numpy as np
    >>> import soundfile as sf
    >>> sf.write('stereo_file.wav', np.random.randn(10, 2), 44100, 'PCM_24')

    """
def blocks(file, blocksize, overlap: int = ..., frames: int = ..., start: int = ..., stop, dtype: str = ..., always_2d: bool = ..., fill_value, out, samplerate, channels, format, subtype, endian, closefd: bool = ...):
    """Return a generator for block-wise reading.

    By default, iteration starts at the beginning and stops at the end
    of the file.  Use *start* to start at a later position and *frames*
    or *stop* to stop earlier.

    If you stop iterating over the generator before it's exhausted,
    the sound file is not closed. This is normally not a problem
    because the file is opened in read-only mode. To close the file
    properly, the generator's ``close()`` method can be called.

    Parameters
    ----------
    file : str or int or file-like object
        The file to read from.  See `SoundFile` for details.
    blocksize : int
        The number of frames to read per block.
        Either this or *out* must be given.
    overlap : int, optional
        The number of frames to rewind between each block.

    Yields
    ------
    `numpy.ndarray` or type(out)
        Blocks of audio data.
        If *out* was given, and the requested frames are not an integer
        multiple of the length of *out*, and no *fill_value* was given,
        the last block will be a smaller view into *out*.

    Other Parameters
    ----------------
    frames, start, stop
        See `read()`.
    dtype : {'float64', 'float32', 'int32', 'int16'}, optional
        See `read()`.
    always_2d, fill_value, out
        See `read()`.
    samplerate, channels, format, subtype, endian, closefd
        See `SoundFile`.

    Examples
    --------
    >>> import soundfile as sf
    >>> for block in sf.blocks('stereo_file.wav', blocksize=1024):
    >>>     pass  # do something with 'block'

    """

class _SoundFileInfo:
    def __init__(self, file, verbose) -> None: ...
    @property
    def _duration_str(self): ...
def info(file, verbose: bool = ...):
    """Returns an object with information about a `SoundFile`.

    Parameters
    ----------
    verbose : bool
        Whether to print additional information.
    """
def available_formats():
    """Return a dictionary of available major formats.

    Examples
    --------
    >>> import soundfile as sf
    >>> sf.available_formats()
    {'FLAC': 'FLAC (FLAC Lossless Audio Codec)',
     'OGG': 'OGG (OGG Container format)',
     'WAV': 'WAV (Microsoft)',
     'AIFF': 'AIFF (Apple/SGI)',
     ...
     'WAVEX': 'WAVEX (Microsoft)',
     'RAW': 'RAW (header-less)',
     'MAT5': 'MAT5 (GNU Octave 2.1 / Matlab 5.0)'}

    """
def available_subtypes(format):
    """Return a dictionary of available subtypes.

    Parameters
    ----------
    format : str
        If given, only compatible subtypes are returned.

    Examples
    --------
    >>> import soundfile as sf
    >>> sf.available_subtypes('FLAC')
    {'PCM_24': 'Signed 24 bit PCM',
     'PCM_16': 'Signed 16 bit PCM',
     'PCM_S8': 'Signed 8 bit PCM'}

    """
def check_format(format, subtype, endian):
    """Check if the combination of format/subtype/endian is valid.

    Examples
    --------
    >>> import soundfile as sf
    >>> sf.check_format('WAV', 'PCM_24')
    True
    >>> sf.check_format('FLAC', 'VORBIS')
    False

    """
def default_subtype(format):
    """Return the default subtype for a given format.

    Examples
    --------
    >>> import soundfile as sf
    >>> sf.default_subtype('WAV')
    'PCM_16'
    >>> sf.default_subtype('MAT5')
    'DOUBLE'

    """

class SoundFile:
    _file: ClassVar[None] = ...
    def __init__(self, file, mode: str = ..., samplerate, channels, subtype, endian, format, closefd: bool = ..., compression_level, bitrate_mode) -> None:
        '''Open a sound file.

        If a file is opened with `mode` ``\'r\'`` (the default) or
        ``\'r+\'``, no sample rate, channels or file format need to be
        given because the information is obtained from the file. An
        exception is the ``\'RAW\'`` data format, which always requires
        these data points.

        File formats consist of three case-insensitive strings:

        * a *major format* which is by default obtained from the
          extension of the file name (if known) and which can be
          forced with the format argument (e.g. ``format=\'WAVEX\'``).
        * a *subtype*, e.g. ``\'PCM_24\'``. Most major formats have a
          default subtype which is used if no subtype is specified.
        * an *endian-ness*, which doesn\'t have to be specified at all in
          most cases.

        A `SoundFile` object is a *context manager*, which means
        if used in a "with" statement, `close()` is automatically
        called when reaching the end of the code block inside the "with"
        statement.

        Parameters
        ----------
        file : str or int or file-like object
            The file to open.  This can be a file name, a file
            descriptor or a Python file object (or a similar object with
            the methods ``read()``/``readinto()``, ``write()``,
            ``seek()`` and ``tell()``).
        mode : {\'r\', \'r+\', \'w\', \'w+\', \'x\', \'x+\'}, optional
            Open mode.  Has to begin with one of these three characters:
            ``\'r\'`` for reading, ``\'w\'`` for writing (truncates *file*)
            or ``\'x\'`` for writing (raises an error if *file* already
            exists).  Additionally, it may contain ``\'+\'`` to open
            *file* for both reading and writing.
            The character ``\'b\'`` for *binary mode* is implied because
            all sound files have to be opened in this mode.
            If *file* is a file descriptor or a file-like object,
            ``\'w\'`` doesn\'t truncate and ``\'x\'`` doesn\'t raise an error.
        samplerate : int
            The sample rate of the file.  If `mode` contains ``\'r\'``,
            this is obtained from the file (except for ``\'RAW\'`` files).
        channels : int
            The number of channels of the file.
            If `mode` contains ``\'r\'``, this is obtained from the file
            (except for ``\'RAW\'`` files).
        subtype : str, sometimes optional
            The subtype of the sound file.  If `mode` contains ``\'r\'``,
            this is obtained from the file (except for ``\'RAW\'``
            files), if not, the default value depends on the selected
            `format` (see `default_subtype()`).
            See `available_subtypes()` for all possible subtypes for
            a given `format`.
        endian : {\'FILE\', \'LITTLE\', \'BIG\', \'CPU\'}, sometimes optional
            The endian-ness of the sound file.  If `mode` contains
            ``\'r\'``, this is obtained from the file (except for
            ``\'RAW\'`` files), if not, the default value is ``\'FILE\'``,
            which is correct in most cases.
        format : str, sometimes optional
            The major format of the sound file.  If `mode` contains
            ``\'r\'``, this is obtained from the file (except for
            ``\'RAW\'`` files), if not, the default value is determined
            from the file extension.  See `available_formats()` for
            all possible values.
        closefd : bool, optional
            Whether to close the file descriptor on `close()`. Only
            applicable if the *file* argument is a file descriptor.
        compression_level : float, optional
            The compression level on \'write()\'. The compression level
            should be between 0.0 (minimum compression level) and 1.0
            (highest compression level).
            See `libsndfile document <https://github.com/libsndfile/libsndfile/blob/c81375f070f3c6764969a738eacded64f53a076e/docs/command.md>`__.
        bitrate_mode : {\'CONSTANT\', \'AVERAGE\', \'VARIABLE\'}, optional
            The bitrate mode on \'write()\'. 
            See `libsndfile document <https://github.com/libsndfile/libsndfile/blob/c81375f070f3c6764969a738eacded64f53a076e/docs/command.md>`__.

        Examples
        --------
        >>> from soundfile import SoundFile

        Open an existing file for reading:

        >>> myfile = SoundFile(\'existing_file.wav\')
        >>> # do something with myfile
        >>> myfile.close()

        Create a new sound file for reading and writing using a with
        statement:

        >>> with SoundFile(\'new_file.wav\', \'x+\', 44100, 2) as myfile:
        >>>     # do something with myfile
        >>>     # ...
        >>>     assert not myfile.closed
        >>>     # myfile.close() is called automatically at the end
        >>> assert myfile.closed

        '''
    def __del__(self) -> None: ...
    def __enter__(self): ...
    def __exit__(self, *args): ...
    def __setattr__(self, name, value):
        """Write text meta-data in the sound file through properties."""
    def __getattr__(self, name):
        """Read text meta-data in the sound file through properties."""
    def __len__(self) -> int: ...
    def __bool__(self) -> bool: ...
    def __nonzero__(self): ...
    def seekable(self):
        """Return True if the file supports seeking."""
    def seek(self, frames, whence: int = ...):
        """Set the read/write position.

        Parameters
        ----------
        frames : int
            The frame index or offset to seek.
        whence : {SEEK_SET, SEEK_CUR, SEEK_END}, optional
            By default (``whence=SEEK_SET``), *frames* are counted from
            the beginning of the file.
            ``whence=SEEK_CUR`` seeks from the current position
            (positive and negative values are allowed for *frames*).
            ``whence=SEEK_END`` seeks from the end (use negative value
            for *frames*).

        Returns
        -------
        int
            The new absolute read/write position in frames.

        Examples
        --------
        >>> from soundfile import SoundFile, SEEK_END
        >>> myfile = SoundFile('stereo_file.wav')

        Seek to the beginning of the file:

        >>> myfile.seek(0)
        0

        Seek to the end of the file:

        >>> myfile.seek(0, SEEK_END)
        44100  # this is the file length

        """
    def tell(self):
        """Return the current read/write position."""
    def read(self, frames: int = ..., dtype: str = ..., always_2d: bool = ..., fill_value, out):
        """Read from the file and return data as NumPy array.

        Reads the given number of frames in the given data format
        starting at the current read/write position.  This advances the
        read/write position by the same number of frames.
        By default, all frames from the current read/write position to
        the end of the file are returned.
        Use `seek()` to move the current read/write position.

        Parameters
        ----------
        frames : int, optional
            The number of frames to read. If ``frames < 0``, the whole
            rest of the file is read.
        dtype : {'float64', 'float32', 'int32', 'int16'}, optional
            Data type of the returned array, by default ``'float64'``.
            Floating point audio data is typically in the range from
            ``-1.0`` to ``1.0``. Integer data is in the range from
            ``-2**15`` to ``2**15-1`` for ``'int16'`` and from
            ``-2**31`` to ``2**31-1`` for ``'int32'``.

            .. note:: Reading int values from a float file will *not*
                scale the data to [-1.0, 1.0). If the file contains
                ``np.array([42.6], dtype='float32')``, you will read
                ``np.array([43], dtype='int32')`` for
                ``dtype='int32'``.

        Returns
        -------
        audiodata : `numpy.ndarray` or type(out)
            A two-dimensional NumPy (frames x channels) array is
            returned. If the sound file has only one channel, a
            one-dimensional array is returned. Use ``always_2d=True``
            to return a two-dimensional array anyway.

            If *out* was specified, it is returned. If *out* has more
            frames than available in the file (or if *frames* is
            smaller than the length of *out*) and no *fill_value* is
            given, then only a part of *out* is overwritten and a view
            containing all valid frames is returned.

        Other Parameters
        ----------------
        always_2d : bool, optional
            By default, reading a mono sound file will return a
            one-dimensional array. With ``always_2d=True``, audio data
            is always returned as a two-dimensional array, even if the
            audio file has only one channel.
        fill_value : float, optional
            If more frames are requested than available in the file,
            the rest of the output is be filled with *fill_value*. If
            *fill_value* is not specified, a smaller array is
            returned.
        out : `numpy.ndarray` or subclass, optional
            If *out* is specified, the data is written into the given
            array instead of creating a new array. In this case, the
            arguments *dtype* and *always_2d* are silently ignored! If
            *frames* is not given, it is obtained from the length of
            *out*.

        Examples
        --------
        >>> from soundfile import SoundFile
        >>> myfile = SoundFile('stereo_file.wav')

        Reading 3 frames from a stereo file:

        >>> myfile.read(3)
        array([[ 0.71329652,  0.06294799],
               [-0.26450912, -0.38874483],
               [ 0.67398441, -0.11516333]])
        >>> myfile.close()

        See Also
        --------
        buffer_read, .write

        """
    def buffer_read(self, frames: int = ..., dtype):
        """Read from the file and return data as buffer object.

        Reads the given number of *frames* in the given data format
        starting at the current read/write position.  This advances the
        read/write position by the same number of frames.
        By default, all frames from the current read/write position to
        the end of the file are returned.
        Use `seek()` to move the current read/write position.

        Parameters
        ----------
        frames : int, optional
            The number of frames to read. If ``frames < 0``, the whole
            rest of the file is read.
        dtype : {'float64', 'float32', 'int32', 'int16'}
            Audio data will be converted to the given data type.

        Returns
        -------
        buffer
            A buffer containing the read data.

        See Also
        --------
        buffer_read_into, .read, buffer_write

        """
    def buffer_read_into(self, buffer, dtype):
        """Read from the file into a given buffer object.

        Fills the given *buffer* with frames in the given data format
        starting at the current read/write position (which can be
        changed with `seek()`) until the buffer is full or the end
        of the file is reached.  This advances the read/write position
        by the number of frames that were read.

        Parameters
        ----------
        buffer : writable buffer
            Audio frames from the file are written to this buffer.
        dtype : {'float64', 'float32', 'int32', 'int16'}
            The data type of *buffer*.

        Returns
        -------
        int
            The number of frames that were read from the file.
            This can be less than the size of *buffer*.
            The rest of the buffer is not filled with meaningful data.

        See Also
        --------
        buffer_read, .read

        """
    def write(self, data):
        """Write audio data from a NumPy array to the file.

        Writes a number of frames at the read/write position to the
        file. This also advances the read/write position by the same
        number of frames and enlarges the file if necessary.

        Note that writing int values to a float file will *not* scale
        the values to [-1.0, 1.0). If you write the value
        ``np.array([42], dtype='int32')``, to a ``subtype='FLOAT'``
        file, the file will then contain ``np.array([42.],
        dtype='float32')``.

        Parameters
        ----------
        data : array_like
            The data to write. Usually two-dimensional (frames x
            channels), but one-dimensional *data* can be used for mono
            files. Only the data types ``'float64'``, ``'float32'``,
            ``'int32'`` and ``'int16'`` are supported.

            .. note:: The data type of *data* does **not** select the
                  data type of the written file. Audio data will be
                  converted to the given *subtype*. Writing int values
                  to a float file will *not* scale the values to
                  [-1.0, 1.0). If you write the value ``np.array([42],
                  dtype='int32')``, to a ``subtype='FLOAT'`` file, the
                  file will then contain ``np.array([42.],
                  dtype='float32')``.

        Examples
        --------
        >>> import numpy as np
        >>> from soundfile import SoundFile
        >>> myfile = SoundFile('stereo_file.wav')

        Write 10 frames of random data to a new file:

        >>> with SoundFile('stereo_file.wav', 'w', 44100, 2, 'PCM_24') as f:
        >>>     f.write(np.random.randn(10, 2))

        See Also
        --------
        buffer_write, .read

        """
    def buffer_write(self, data, dtype):
        """Write audio data from a buffer/bytes object to the file.

        Writes the contents of *data* to the file at the current
        read/write position.
        This also advances the read/write position by the number of
        frames that were written and enlarges the file if necessary.

        Parameters
        ----------
        data : buffer or bytes
            A buffer or bytes object containing the audio data to be
            written.
        dtype : {'float64', 'float32', 'int32', 'int16'}
            The data type of the audio data stored in *data*.

        See Also
        --------
        .write, buffer_read

        """
    def blocks(self, blocksize, overlap: int = ..., frames: int = ..., dtype: str = ..., always_2d: bool = ..., fill_value, out):
        """Return a generator for block-wise reading.

        By default, the generator yields blocks of the given
        *blocksize* (using a given *overlap*) until the end of the file
        is reached; *frames* can be used to stop earlier.

        Parameters
        ----------
        blocksize : int
            The number of frames to read per block. Either this or *out*
            must be given.
        overlap : int, optional
            The number of frames to rewind between each block.
        frames : int, optional
            The number of frames to read.
            If ``frames < 0``, the file is read until the end.
        dtype : {'float64', 'float32', 'int32', 'int16'}, optional
            See `read()`.

        Yields
        ------
        `numpy.ndarray` or type(out)
            Blocks of audio data.
            If *out* was given, and the requested frames are not an
            integer multiple of the length of *out*, and no
            *fill_value* was given, the last block will be a smaller
            view into *out*.


        Other Parameters
        ----------------
        always_2d, fill_value, out
            See `read()`.
        fill_value : float, optional
            See `read()`.
        out : `numpy.ndarray` or subclass, optional
            If *out* is specified, the data is written into the given
            array instead of creating a new array. In this case, the
            arguments *dtype* and *always_2d* are silently ignored!

        Examples
        --------
        >>> from soundfile import SoundFile
        >>> with SoundFile('stereo_file.wav') as f:
        >>>     for block in f.blocks(blocksize=1024):
        >>>         pass  # do something with 'block'

        """
    def truncate(self, frames):
        """Truncate the file to a given number of frames.

        After this command, the read/write position will be at the new
        end of the file.

        Parameters
        ----------
        frames : int, optional
            Only the data before *frames* is kept, the rest is deleted.
            If not specified, the current read/write position is used.

        """
    def flush(self):
        """Write unwritten data to the file system.

        Data written with `write()` is not immediately written to
        the file system but buffered in memory to be written at a later
        time.  Calling `flush()` makes sure that all changes are
        actually written to the file system.

        This has no effect on files opened in read-only mode.

        """
    def close(self):
        """Close the file.  Can be called multiple times."""
    def _open(self, file, mode_int, closefd):
        """Call the appropriate sf_open*() function from libsndfile."""
    def _init_virtual_io(self, file):
        """Initialize callback functions for sf_open_virtual()."""
    def _getAttributeNames(self):
        """Return all attributes used in __setattr__ and __getattr__.

        This is useful for auto-completion (e.g. IPython).

        """
    def _check_if_closed(self):
        """Check if the file is closed and raise an error if it is.

        This should be used in every method that uses self._file.

        """
    def _check_frames(self, frames, fill_value):
        """Reduce frames to no more than are available in the file."""
    def _check_buffer(self, data, ctype):
        """Convert buffer to cdata and check for valid size."""
    def _create_empty_array(self, frames, always_2d, dtype):
        """Create an empty array with appropriate shape."""
    def _check_dtype(self, dtype):
        """Check if dtype string is valid and return ctype string."""
    def _array_io(self, action, array, frames):
        """Check array and call low-level IO function."""
    def _cdata_io(self, action, data, ctype, frames):
        """Call one of libsndfile's read/write functions."""
    def _update_frames(self, written):
        """Update self.frames after writing."""
    def _prepare_read(self, start, stop, frames):
        """Seek to start frame and calculate length."""
    def copy_metadata(self):
        """Get all metadata present in this SoundFile

        Returns
        -------

        metadata: dict[str, str]
            A dict with all metadata. Possible keys are: 'title', 'copyright',
            'software', 'artist', 'comment', 'date', 'album', 'license',
            'tracknumber' and 'genre'.
        """
    def _set_bitrate_mode(self, bitrate_mode):
        """Call libsndfile's set bitrate mode function."""
    def _set_compression_level(self, compression_level):
        """Call libsndfile's set compression level function."""
    @property
    def name(self): ...
    @property
    def mode(self): ...
    @property
    def samplerate(self): ...
    @property
    def frames(self): ...
    @property
    def channels(self): ...
    @property
    def format(self): ...
    @property
    def subtype(self): ...
    @property
    def endian(self): ...
    @property
    def format_info(self): ...
    @property
    def subtype_info(self): ...
    @property
    def sections(self): ...
    @property
    def closed(self): ...
    @property
    def _errorcode(self): ...
    @property
    def compression_level(self): ...
    @property
    def bitrate_mode(self): ...
    @property
    def extra_info(self): ...
def _error_check(err, prefix: str = ...):
    """Raise LibsndfileError if there is an error."""
def _format_int(format, subtype, endian):
    """Return numeric ID for given format|subtype|endian combo."""
def _check_mode(mode):
    """Check if mode is valid and return its integer representation."""
def _create_info_struct(file, mode, samplerate, channels, format, subtype, endian):
    """Check arguments and create SF_INFO struct."""
def _get_format_from_filename(file, mode):
    """Return a format string obtained from file (or file.name).

    If file already exists (= read mode), an empty string is returned on
    error.  If not, an exception is raised.
    The return type will always be str or unicode (even if
    file/file.name is a bytes object).

    """
def _format_str(format_int):
    """Return the string representation of a given numeric format."""
def _format_info(format_int, format_flag: int = ...):
    """Return the ID and short description of a given format."""
def _available_formats_helper(count_flag, format_flag):
    """Helper for available_formats() and available_subtypes()."""
def _check_format(format_str):
    """Check if `format_str` is valid and return format ID."""
def _has_virtual_io_attrs(file, mode_int):
    """Check if file has all the necessary attributes for virtual IO."""

class SoundFileError(Exception): ...
class SoundFileRuntimeError(SoundFileError, RuntimeError): ...

class LibsndfileError(SoundFileRuntimeError):
    def __init__(self, code, prefix: str = ...) -> None: ...
    @property
    def error_string(self): ...



__all__ = ['get_filter', 'clear_cache', 'sinc_window']

def sinc_window(num_zeros: int = ..., precision: int = ..., window, rolloff: float = ...):
    """Construct a windowed sinc interpolation filter

    Parameters
    ----------
    num_zeros : int > 0
        The number of zero-crossings to retain in the sinc filter
    precision : int > 0
        The number of filter coefficients to retain for each zero-crossing
    window : callable
        The window function.  By default, uses a Hann window.
    rolloff : float > 0
        The roll-off frequency (as a fraction of nyquist)

    Returns
    -------
    interp_window: np.ndarray [shape=(num_zeros * num_table + 1)]
        The interpolation window (right-hand side)
    num_bits: int
        The number of bits of precision to use in the filter table
    rolloff : float > 0
        The roll-off frequency of the filter, as a fraction of Nyquist

    Raises
    ------
    TypeError
        if `window` is not callable or `None`
    ValueError
        if `num_zeros < 1`, `precision < 1`,
        or `rolloff` is outside the range `(0, 1]`.

    Examples
    --------
    >>> import scipy, scipy.signal
    >>> import resampy
    >>> np.set_printoptions(threshold=5, suppress=False)
    >>> # A filter with 10 zero-crossings, 32 samples per crossing, and a
    >>> # Hann window for tapering.
    >>> halfwin, prec, rolloff = resampy.filters.sinc_window(num_zeros=10, precision=5,
    ...                                                      window=scipy.signal.windows.hann)
    >>> halfwin
    array([  9.450e-01,   9.436e-01, ...,  -7.455e-07,  -0.000e+00])
    >>> prec
    32
    >>> rolloff
    0.945

    >>> # Or using sinc-window filter construction directly in resample
    >>> y = resampy.resample(x, sr_orig, sr_new, filter='sinc_window',
    ...                      num_zeros=10, precision=5,
    ...                      window=scipy.signal.windows.hann)      # doctest: +SKIP
    """
def get_filter(name_or_function, **kwargs):
    """Retrieve a window given its name or function handle.

    Parameters
    ----------
    name_or_function : str or callable
        If a function, returns `name_or_function(**kwargs)`.

        If a string, and it matches the name of one of the defined
        filter functions, the corresponding function is called with `**kwargs`.

        If a string, and it matches the name of a pre-computed filter,
        the corresponding filter is retrieved, and kwargs is ignored.

        Valid pre-computed filter names are:
            - 'kaiser_fast'
            - 'kaiser_best'

    **kwargs
        Additional keyword arguments passed to `name_or_function` (if callable)

    Returns
    -------
    half_window : np.ndarray
        The right wing of the interpolation filter
    precision : int > 0
        The number of samples between zero-crossings of the filter
    rolloff : float > 0
        The roll-off frequency of the filter as a fraction of Nyquist

    Raises
    ------
    NotImplementedError
        If `name_or_function` cannot be found as a filter.
    """
def clear_cache():
    """Clear the filter cache.

    Calling this function will ensure that packaged filters are reloaded
    upon the next usage.
    """



"""
Provides utilities for string extraction from nested data structures
and merges multiple dictionaries containing lists into one dictionary.
"""

from numpy.typing import NDArray
from typing import Any
import more_itertools
import numpy
import python_minifier
import re as regex

def autoDecodingRLE(arrayTarget: NDArray[numpy.integer[Any]], addSpaces: bool = False, axisOfOperation: int | None = None) -> str:
	"""Special case, range, start=0"""
	if axisOfOperation is None:
		axisOfOperation = 0
	def sliceNDArrayToNestedLists(arraySlice: NDArray[numpy.integer[Any]], axisOfOperation: int) -> Any:
		if isinstance(arraySlice, numpy.ndarray) and arraySlice.ndim > 1:
			if (axisOfOperation >= arraySlice.ndim):
				axisOfOperation = -1
			elif abs(axisOfOperation) > arraySlice.ndim:
				axisOfOperation = 0
			return [sliceNDArrayToNestedLists(arraySlice[index], axisOfOperation) for index in range(arraySlice.shape[axisOfOperation])]
		elif isinstance(arraySlice, numpy.ndarray) and arraySlice.ndim == 1:
			arraySliceAsList = []
			for seriesGrouped in more_itertools.consecutive_groups(arraySlice.tolist()):
				ImaSerious = list(seriesGrouped)
				ImaRange = [range(ImaSerious[0], ImaSerious[-1] + 1)]
				lengthAsList = addSpaces*(len(ImaSerious)-1) + len(python_minifier.minify(str(ImaSerious))) # brackets are proxies for commas
				lengthAsRange = addSpaces*1 + len('*') + len(python_minifier.minify(str(ImaRange)))
				if lengthAsRange < lengthAsList:
					arraySliceAsList += ImaRange
				else:
					arraySliceAsList += ImaSerious
			COPYarraySliceAsList = arraySliceAsList.copy()
			arraySliceAsList = []
			for malkovichGrouped in more_itertools.run_length.encode(COPYarraySliceAsList):
				lengthMalkovich = malkovichGrouped[-1]
				malkovichAsList = list(more_itertools.run_length.decode([malkovichGrouped]))
				lengthAsList = addSpaces*(len(malkovichAsList)-1) + len(python_minifier.minify(str(malkovichAsList))) # brackets are proxies for commas
				malkovichMalkovich = f"[{malkovichGrouped[0]}]*{lengthMalkovich}"
				lengthAsMalkovich = len(python_minifier.minify(malkovichMalkovich))
				if lengthAsMalkovich < lengthAsList:
					arraySliceAsList.append(malkovichGrouped)
				else:
					arraySliceAsList += malkovichAsList
			return arraySliceAsList
		return arraySlice

	arrayAsNestedLists = sliceNDArrayToNestedLists(arrayTarget, axisOfOperation)

	arrayAsStr = python_minifier.minify(str(arrayAsNestedLists))

	for insanity in range(2):
		joinAheadComma = regex.compile("(?<!rang)(?P<joinAhead>,)\\((?P<malkovich>\\d+),(?P<multiple>\\d+)\\)(?P<joinBehind>])")
		joinAheadCommaReplace = "]+[\\g<malkovich>]*\\g<multiple>"
		arrayAsStr = joinAheadComma.sub(joinAheadCommaReplace, arrayAsStr)

		joinBehindComma = regex.compile("(?<!rang)(?P<joinAhead>\\[|^.)\\((?P<malkovich>\\d+),(?P<multiple>\\d+)\\)(?P<joinBehind>,)")
		joinBehindCommaReplace = "[\\g<malkovich>]*\\g<multiple>+["
		arrayAsStr = joinBehindComma.sub(joinBehindCommaReplace, arrayAsStr)

		joinAheadBracket = regex.compile("(?<!rang)(?P<joinAhead>\\[)\\((?P<malkovich>\\d+),(?P<multiple>\\d+)\\)(?P<joinBehind>])")
		joinAheadBracketReplace = "[\\g<malkovich>]*\\g<multiple>"
		arrayAsStr = joinAheadBracket.sub(joinAheadBracketReplace, arrayAsStr)

		joinBothCommas = regex.compile("(?<!rang)(?P<joinAhead>,)\\((?P<malkovich>\\d+),(?P<multiple>\\d+)\\)(?P<joinBehind>,)")
		joinBothCommasReplace = "]+[\\g<malkovich>]*\\g<multiple>+["
		arrayAsStr = joinBothCommas.sub(joinBothCommasReplace, arrayAsStr)

	arrayAsStr = arrayAsStr.replace('range', '*range')

	return arrayAsStr

def stringItUp(*scrapPile: Any) -> list[str]:
	"""
	Convert, if possible, every element in the input data structure to a string. Order is not preserved or readily predictable.

	Parameters:
		*scrapPile: One or more data structures to unpack and convert to strings.
	Returns:
		listStrungUp: A list of string versions of all convertible elements.
	"""
	scrap = None
	listStrungUp: list[str] = []

	def drill(KitKat: Any) -> None:
		if isinstance(KitKat, str):
			listStrungUp.append(KitKat)
		elif isinstance(KitKat, (bool, bytearray, bytes, complex, float, int, memoryview, type(None))):
			listStrungUp.append(str(KitKat))
		elif isinstance(KitKat, dict):
			for broken, piece in KitKat.items():
				drill(broken)
				drill(piece)
		elif isinstance(KitKat, (frozenset, list, range, set, tuple)):
			for kit in KitKat:
				drill(kit)
		elif hasattr(KitKat, '__iter__'): # Unpack other iterables
			for kat in KitKat:
				drill(kat)
		else:
			try:
				sharingIsCaring = KitKat.__str__()
				listStrungUp.append(sharingIsCaring)
			except AttributeError:
				pass
			except TypeError: # "The error traceback provided indicates that there is an issue when calling the __str__ method on an object that does not have this method properly defined, leading to a TypeError."
				pass
			except:
				print(f"\nWoah! I received '{repr(KitKat)}'.\nTheir report card says, 'Plays well with others: Needs improvement.'\n")
				raise
	try:
		for scrap in scrapPile:
			drill(scrap)
	except RecursionError:
		listStrungUp.append(repr(scrap))
	return listStrungUp

def updateExtendPolishDictionaryLists(*dictionaryLists: dict[str, list[Any] | set[Any] | tuple[Any, ...]], destroyDuplicates: bool = False, reorderLists: bool = False, killErroneousDataTypes: bool = False) -> dict[str, list[Any]]:
	"""
	Merges multiple dictionaries containing lists into a single dictionary, with options to handle duplicates,
	list ordering, and erroneous data types.

	Parameters:
		*dictionaryLists: Variable number of dictionaries to be merged. If only one dictionary is passed, it will be processed based on the provided options.
		destroyDuplicates (False): If True, removes duplicate elements from the lists. Defaults to False.
		reorderLists (False): If True, sorts the lists. Defaults to False.
		killErroneousDataTypes (False): If True, skips dictionary keys or dictionary values that cause a TypeError during merging. Defaults to False.
	Returns:
		ePluribusUnum: A single dictionary with merged lists based on the provided options. If only one dictionary is passed,
		it will be cleaned up based on the options.
	Note:
		The returned value, `ePluribusUnum`, is a so-called primitive dictionary (`typing.Dict`). Furthermore, every dictionary key is a so-called primitive string (cf. `str()`) and every dictionary value is a so-called primitive list (`typing.List`). If `dictionaryLists` has other data types, the data types will not be preserved. That could have unexpected consequences. Conversion from the original data type to a `typing.List`, for example, may not preserve the order even if you want the order to be preserved.
	"""

	ePluribusUnum: dict[str, list[Any]] = {}

	for dictionaryListTarget in dictionaryLists:
		for keyName, keyValue in dictionaryListTarget.items():
			try:
				ImaStr = str(keyName)
				ImaList = list(keyValue)
				ePluribusUnum.setdefault(ImaStr, []).extend(ImaList)
			except TypeError:
				if killErroneousDataTypes:
					continue
				else:
					raise

	if destroyDuplicates:
		for ImaStr, ImaList in ePluribusUnum.items():
			ePluribusUnum[ImaStr] = list(dict.fromkeys(ImaList))
	if reorderLists:
		for ImaStr, ImaList in ePluribusUnum.items():
			ePluribusUnum[ImaStr] = sorted(ImaList)

	return ePluribusUnum



"""
Provides utilities for reading, writing, and resampling audio waveforms.
"""
from numpy import complex64, complexfloating, dtype, float32, floating, ndarray, integer
from numpy.typing import NDArray
from scipy.signal import ShortTimeFFT
from scipy.signal._short_time_fft import PAD_TYPE, FFT_MODE_TYPE
from typing import Any, BinaryIO, Literal, TypedDict, cast, overload
from collections.abc import Sequence
from Z0Z_tools import halfsine, makeDirsSafely
import functools
import io
import math
import numpy
import numpy.typing
import os
import resampy
import soundfile

class ParametersSTFT(TypedDict, total=False):
	padding: PAD_TYPE
	axis: int

class ParametersShortTimeFFT(TypedDict, total=False):
	fft_mode: FFT_MODE_TYPE
	scale_to: Literal['magnitude', 'psd']

class ParametersUniversal(TypedDict):
	lengthFFT: int
	lengthHop: int
	lengthWindowingFunction: int
	sampleRate: float
	windowingFunction: ndarray[tuple[int], dtype[floating[Any]]]

class WaveformMetadata(TypedDict):
	pathFilename: str
	lengthWaveform: int
	samplesLeading: int
	samplesTrailing: int

# TODO how should I handle these?
parametersShortTimeFFTUniversal: ParametersShortTimeFFT = {'fft_mode': 'onesided'}
parametersSTFTUniversal: ParametersSTFT = {'padding': 'even', 'axis': -1}

lengthWindowingFunctionDEFAULT = 1024
windowingFunctionCallableDEFAULT = halfsine
parametersDEFAULT = ParametersUniversal (
	lengthFFT=2048,
	lengthHop=512,
	lengthWindowingFunction=lengthWindowingFunctionDEFAULT,
	sampleRate=44100,
	windowingFunction=windowingFunctionCallableDEFAULT(lengthWindowingFunctionDEFAULT),
)

# No, I don't know how to implement this, but I might learn how to do it later.
# If you know how, you can help. :D
parametersUniversal = {}

windowingFunctionCallableUniversal = windowingFunctionCallableDEFAULT
if not parametersUniversal:
	parametersUniversal = parametersDEFAULT

def getWaveformMetadata(listPathFilenames: Sequence[str | os.PathLike[str]], sampleRate: float) -> dict[int, WaveformMetadata]:
	axisTime: int = -1
	dictionaryWaveformMetadata: dict[int, WaveformMetadata] = {}
	for index, pathFilename in enumerate(listPathFilenames):
		lengthWaveform = readAudioFile(pathFilename, sampleRate).shape[axisTime]
		dictionaryWaveformMetadata[index] = WaveformMetadata(
			pathFilename = str(pathFilename),
			lengthWaveform = lengthWaveform,
			samplesLeading = 0,
			samplesTrailing = 0,
		)
	return dictionaryWaveformMetadata

def readAudioFile(pathFilename: str | os.PathLike[Any] | BinaryIO, sampleRate: float | None = None) -> ndarray[tuple[Literal[2], int], dtype[float32]]:
	"""
	Reads an audio file and returns its data as a NumPy array. Mono is always converted to stereo.

	Parameters:
		pathFilename: The path to the audio file.
		sampleRate (44100): The sample rate of the returned waveform. Defaults to 44100.

	Returns:
		waveform: The audio data in an array shaped (channels, samples).
	"""
	if sampleRate is None:
		sampleRate = parametersUniversal['sampleRate']
	try:
		with soundfile.SoundFile(pathFilename) as readSoundFile:
			sampleRateSource: int = readSoundFile.samplerate
			waveform: NDArray[float32] = readSoundFile.read(dtype='float32', always_2d=True).astype(float32)
			# GitHub #3 Implement semantic axes for audio data
			axisTime = 0; axisChannels = 1
			waveform = resampleWaveform(waveform, sampleRateDesired=sampleRate, sampleRateSource=sampleRateSource, axisTime=axisTime)
			if waveform.shape[axisChannels] == 1:
				waveform = numpy.repeat(waveform, 2, axis=axisChannels)
			return numpy.transpose(waveform, axes=(axisChannels, axisTime))
	except soundfile.LibsndfileError as ERRORmessage:
		if 'System error' in str(ERRORmessage):
			raise FileNotFoundError(f"File not found: {pathFilename}") from ERRORmessage
		else:
			raise

def resampleWaveform(waveform: NDArray[floating[Any]], sampleRateDesired: float, sampleRateSource: float, axisTime: int = -1) -> NDArray[float32]:
	"""
	Resamples the waveform to the desired sample rate using resampy.

	Parameters:
		waveform: The input audio data.
		sampleRateDesired: The desired sample rate.
		sampleRateSource: The original sample rate of the waveform.

	Returns:
		waveformResampled: The resampled waveform.
	"""
	if sampleRateSource != sampleRateDesired:
		sampleRateDesired = round(sampleRateDesired)
		sampleRateSource = round(sampleRateSource)
		waveformResampled: NDArray[float32] = resampy.resample(waveform, sampleRateSource, sampleRateDesired, axis=axisTime)
		return waveformResampled
	else:
		return waveform

def loadWaveforms(listPathFilenames: Sequence[str | os.PathLike[str]], sampleRateTarget: float | None = None) -> ndarray[tuple[int, int, int], dtype[float32]]:
	"""
	Load a list of audio files into a single array.

	Parameters:
		listPathFilenames: List of file paths to the audio files.
		sampleRate (44100): Target sample rate for the waveforms; the function will resample if necessary. Defaults to 44100.
	Returns:
		arrayWaveforms: A single NumPy array of shape (countChannels, lengthWaveformMaximum, countWaveforms)
	"""
	if sampleRateTarget is None:
		sampleRateTarget = parametersUniversal['sampleRate']

	# GitHub #3 Implement semantic axes for audio data
	axisOrderMapping: dict[str, int] = {'indexingAxis': -1, 'axisTime': -2, 'axisChannels': 0}
	axesSizes: dict[str, int] = {keyName: 1 for keyName in axisOrderMapping.keys()}
	countAxes: int = len(axisOrderMapping)
	listShapeIndexToSize: list[int] = [9001] * countAxes

	countWaveforms: int = len(listPathFilenames)
	axesSizes['indexingAxis'] = countWaveforms
	countChannels: int = 2
	axesSizes['axisChannels'] = countChannels

	axisTime: int = -1
	dictionaryWaveformMetadata = getWaveformMetadata(listPathFilenames, sampleRateTarget)

	samplesTotalMaximum = max([entry['lengthWaveform'] + entry['samplesLeading'] + entry['samplesTrailing'] for entry in dictionaryWaveformMetadata.values()])
	axesSizes['axisTime'] = samplesTotalMaximum

	for keyName, axisSize in axesSizes.items():
		axisNormalized: int = (axisOrderMapping[keyName] + countAxes) % countAxes
		listShapeIndexToSize[axisNormalized] = axisSize
	tupleShapeArray = cast(tuple[int, int, int], tuple(listShapeIndexToSize))

	# `numpy.zeros` so that shorter waveforms are safely padded with zeros
	arrayWaveforms: ndarray[tuple[int, int, int], dtype[float32]] = numpy.zeros(tupleShapeArray, dtype=float32)

	for index, metadata in dictionaryWaveformMetadata.items():
		waveform = readAudioFile(metadata['pathFilename'], sampleRateTarget)
		samplesTrailing = metadata['lengthWaveform'] + metadata['samplesLeading'] - samplesTotalMaximum
		if samplesTrailing == 0:
			samplesTrailing = None
		# padding logic, entry['samplesLeading'] + entry['samplesTrailing'], goes here
		"""TODO padding logic; thoughts about the following statement.
		If my goal were to reduce the chances of an exception, especially a broadcast exception, in the next statement,
		then I would slice the insertion like this:
		`arrayWaveforms[:, metadata['samplesLeading'] : metadata['samplesLeading'] + waveform.shape[axisTime], index] = waveform`
		That would ensure that the exact length of the waveform that was just loaded would be used as the length of the slice.
		In rare cases, samplesLeading+lengthWaveform might be too long, which would cause an exception. And if that were to happen,
		the actual cause of the problem would almost certainly be upstream from here: that would make it harder to troubleshoot. Also,
		preallocating arrayWaveforms assumes precision; padding multiple waveforms with leading and or trailing requires precision:
		precision to the exact sample. The following statement doesn't affect precision, but if there is a precision problem,
		then the following statement could be affected by it and raise an exception: fail early. If the padding logic is implemented,
		I think the best form of the state will be:
		`arrayWaveforms[:, metadata['samplesLeading']:-metadata['samplesTrailing'], index] = waveform`
		If any of the upstream calculations is wrong, then the above statement would likely induce a broadcasting exception: fail early.

		I don't have padding logic right now, but I constructed the tedious `samplesTrailing` logic because it is slightly more likely
		to fail early than the alternative and because if I were to implement the padding logic, during refactoring, I would be more
		likely to notice this tedious logic and remember to implement the idea above.
		"""
		arrayWaveforms[:, metadata['samplesLeading']:samplesTrailing, index] = waveform

	return arrayWaveforms

def writeWAV(pathFilename: str | os.PathLike[Any] | io.IOBase, waveform: ndarray[tuple[int, ...], dtype[floating[Any] | integer[Any]]], sampleRate: float | None = None) -> None:
	"""
	Writes a waveform to a WAV file.

	Parameters:
		pathFilename: The path and filename where the WAV file will be saved.
		waveform: The waveform data to be written to the WAV file. The waveform should be in the shape (channels, samples) or (samples,).
		sampleRate (44100): The sample rate of the waveform. Defaults to 44100 Hz.

	Returns:
		None:

	### Note well
		The function overwrites existing files without prompting or informing the user.

	Notes
		All files are saved as 32-bit float.
		The function will attempt to create the directory structure, if applicable.
	"""
	if sampleRate is None:
		sampleRate = parametersUniversal['sampleRate']
	makeDirsSafely(pathFilename)
	soundfile.write(file=pathFilename, data=waveform.T, samplerate=sampleRate, subtype='FLOAT', format='WAV')

@overload # stft, one waveform
def stft(arrayTarget: ndarray[tuple[int, int], dtype[floating[Any] | integer[Any]]]
		, *
		, sampleRate: float | None = None
		, lengthHop: int | None = None
		, windowingFunction: ndarray[tuple[int], dtype[floating[Any]]] | None = None
		, lengthWindowingFunction: int | None = None
		, lengthFFT: int | None = None
		, inverse: Literal[False] = False
		, lengthWaveform: None = None
		, indexingAxis: Literal[None] = None
		) -> ndarray[tuple[int, int, int], dtype[complexfloating[Any, Any]]]: ...

@overload # stft, array of waveforms
def stft(arrayTarget: ndarray[tuple[int, int, int], dtype[floating[Any] | integer[Any]]]
		, *
		, sampleRate: float | None = None
		, lengthHop: int | None = None
		, windowingFunction: ndarray[tuple[int], dtype[floating[Any]]] | None = None
		, lengthWindowingFunction: int | None = None
		, lengthFFT: int | None = None
		, inverse: Literal[False] = False
		, lengthWaveform: None = None
		, indexingAxis: int = -1
		) -> ndarray[tuple[int, int, int, int], dtype[complexfloating[Any, Any]]]: ...

@overload # istft, one spectrogram
def stft(arrayTarget: ndarray[tuple[int, int, int], dtype[complexfloating[Any, Any] | floating[Any]]]
		, *
		, sampleRate: float | None = None
		, lengthHop: int | None = None
		, windowingFunction: ndarray[tuple[int], dtype[floating[Any]]] | None = None
		, lengthWindowingFunction: int | None = None
		, lengthFFT: int | None = None
		, inverse: Literal[True]
		, lengthWaveform: int
		, indexingAxis: Literal[None] = None
		) -> ndarray[tuple[int, int], dtype[floating[Any]]]: ...

@overload # istft, array of spectrograms
def stft(arrayTarget: ndarray[tuple[int, int, int, int], dtype[complexfloating[Any, Any]]]
		, *
		, sampleRate: float | None = None
		, lengthHop: int | None = None
		, windowingFunction: ndarray[tuple[int], dtype[floating[Any]]] | None = None
		, lengthWindowingFunction: int | None = None
		, lengthFFT: int | None = None
		, inverse: Literal[True]
		, lengthWaveform: int
		, indexingAxis: int = -1
		) -> ndarray[tuple[int, int, int], dtype[floating[Any]]]: ...

def stft(arrayTarget: (ndarray[tuple[int, int], 		   dtype[floating[Any] | integer[Any]]]
						 |   ndarray[tuple[int, int, int], 	   dtype[floating[Any] | integer[Any]]]
						 |   ndarray[tuple[int, int, int], 	   dtype[complexfloating[Any, Any] | floating[Any]]]
						 |   ndarray[tuple[int, int, int, int], dtype[complexfloating[Any, Any]]])
		, *
		, sampleRate: float | None = None
		, lengthHop: int | None = None
		, windowingFunction: ndarray[tuple[int], dtype[floating[Any]]] | None = None
		, lengthWindowingFunction: int | None = None
		, lengthFFT: int | None = None
		, inverse: bool = False
		, lengthWaveform: int | None = None
		, indexingAxis: int | None = None
		) -> (ndarray[tuple[int, int], 		  dtype[floating[Any]]]
				 |  ndarray[tuple[int, int, int], 	  dtype[floating[Any]]]
				 |  ndarray[tuple[int, int, int], 	  dtype[complexfloating[Any, Any]]]
				 |  ndarray[tuple[int, int, int, int], dtype[complexfloating[Any, Any]]]):
	"""
	Short-Time Fourier Transform with unified interface for forward and inverse transforms.

	Parameters:
		arrayTarget: Input array for transformation.
		sampleRate (44100): Sample rate of the signal.
		lengthHop (512): Number of samples between successive frames.
		windowingFunction (halfsine): Windowing function array. Defaults to halfsine if None.
		lengthWindowingFunction (1024): Length of the windowing function. Used if windowingFunction is None.
		lengthFFT (2048*): Length of the FFT. Defaults to 2048 or the next power of 2 >= lengthWindowingFunction.
		inverse (False): Whether to perform inverse transform.
		lengthWaveform: Required output length for inverse transform.
		indexingAxis (None, -1): Axis containing multiple signals to transform.

	Returns:
		arrayTransformed: The transformed signal(s).
	"""
	if sampleRate is None: sampleRate = parametersUniversal['sampleRate']
	if lengthHop is None: lengthHop = parametersUniversal['lengthHop']

	if windowingFunction is None:
		if lengthWindowingFunction is not None and windowingFunctionCallableUniversal:
			windowingFunction = windowingFunctionCallableUniversal(lengthWindowingFunction)
		else:
			windowingFunction = parametersUniversal['windowingFunction']
		if lengthFFT is None:
			lengthFFTSherpa = parametersUniversal['lengthFFT']
			if lengthFFTSherpa >= windowingFunction.size:
				lengthFFT = lengthFFTSherpa

	if lengthFFT is None:
		lengthWindowingFunction = windowingFunction.size
		lengthFFT = 2 ** math.ceil(math.log2(lengthWindowingFunction))

	if inverse and not lengthWaveform:
		raise ValueError("lengthWaveform must be specified for inverse transform")

	stftWorkhorse = ShortTimeFFT(win=windowingFunction, hop=lengthHop, fs=sampleRate, mfft=lengthFFT, **parametersShortTimeFFTUniversal)

	@overload
	def doTransformation(arrayInput: 	   ndarray[tuple[int, int, int], 	   dtype[complexfloating[Any, Any] | floating[Any]]]
										, lengthWaveform: int, inverse: Literal[True] = True
								) -> 	   ndarray[tuple[int, int], 		   dtype[floating[Any]]]: ...
	@overload
	def doTransformation(arrayInput: 	   ndarray[tuple[int, int], 		   dtype[floating[Any] | integer[Any]]]
										, lengthWaveform: Literal[None] = None, inverse: Literal[False] = False
								) -> 	   ndarray[tuple[int, int, int],	   dtype[complexfloating[Any, Any]]]: ...
	def doTransformation(arrayInput: (ndarray[tuple[int, int], 		   dtype[floating[Any] | integer[Any]]]
										 |  ndarray[tuple[int, int, int], 	   dtype[complexfloating[Any, Any] | floating[Any]]])
										, lengthWaveform: int | None = lengthWaveform, inverse: bool | None = inverse
								) -> (ndarray[tuple[int, int], 		   dtype[floating[Any]]]
										 |  ndarray[tuple[int, int, int], 	   dtype[complexfloating[Any, Any]]]):
		if inverse:
			return stftWorkhorse.istft(S=arrayInput, k1=lengthWaveform)
		return stftWorkhorse.stft(x=arrayInput, **parametersSTFTUniversal)

	# No overloads for "doTransformation" match the provided arguments Pylance(reportCallIssue)
	# Pylance, why do you hate me?
	if indexingAxis is None:
		return doTransformation(arrayTarget, inverse=inverse, lengthWaveform=lengthWaveform) # type: ignore

	arrayTARGET = numpy.moveaxis(arrayTarget, indexingAxis, -1)
	index = 0
	arrayTransformed = numpy.tile(doTransformation(arrayTARGET[..., index], inverse, lengthWaveform)[..., numpy.newaxis], arrayTARGET.shape[-1]) # type: ignore

	for index in range(1, arrayTARGET.shape[-1]):
		arrayTransformed[..., index] = doTransformation(arrayTARGET[..., index], inverse, lengthWaveform) # type: ignore

	return numpy.moveaxis(arrayTransformed, -1, indexingAxis)

def loadSpectrograms(listPathFilenames: Sequence[str | os.PathLike[str]]
					, sampleRateTarget: float | None = None
					, **parametersSTFT: Any
					) -> tuple[ndarray[tuple[int, int, int, int], dtype[complex64]], dict[int, WaveformMetadata]]:
	"""
	Load spectrograms from audio files.

	Parameters:
		listPathFilenames: A list of WAV path and filenames.
		sampleRateTarget (44100): The target sample rate. If necessary, a file will be resampled to the target sample rate. Defaults to 44100.
		**parametersSTFT: Keyword-parameters for the Short-Time Fourier Transform, see `stft`.

	Returns:
		tupleSpectrogramsLengthsWaveform: A tuple containing the array of spectrograms and a list of metadata dictionaries for each spectrogram.
	"""
	if sampleRateTarget is None:
		sampleRateTarget = parametersUniversal['sampleRate']

	# TODO padding logic
	dictionaryWaveformMetadata = getWaveformMetadata(listPathFilenames, sampleRateTarget)

	samplesTotalMaximum = max([entry['lengthWaveform'] + entry['samplesLeading'] + entry['samplesTrailing'] for entry in dictionaryWaveformMetadata.values()])

	countChannels = 2
	spectrogramArchetype: ndarray[tuple[int, int, int], dtype[complex64]] = stft(numpy.zeros(shape=(countChannels, samplesTotalMaximum), dtype=float32), sampleRate=sampleRateTarget, **parametersSTFT)
	arraySpectrograms = numpy.zeros(shape=(*spectrogramArchetype.shape, len(dictionaryWaveformMetadata)), dtype=numpy.complex64)

	for index, metadata in dictionaryWaveformMetadata.items():
		waveform = readAudioFile(metadata['pathFilename'], sampleRateTarget)
		# padding logic, entry['samplesLeading'] + entry['samplesTrailing'], goes here
		arraySpectrograms[..., index] = stft(waveform, sampleRate=sampleRateTarget, **parametersSTFT)

	return arraySpectrograms, dictionaryWaveformMetadata

def spectrogramToWAV( spectrogram: ndarray[tuple[int, int, int], dtype[complexfloating[Any, Any] | floating[Any]]]
					, pathFilename: str | os.PathLike[Any] | io.IOBase
					, lengthWaveform: int
					, sampleRate: float | None = None
					, **parametersSTFT: Any
					) -> None:
	"""
	Writes a complex spectrogram to a WAV file.

	Parameters:
		spectrogram: The complex spectrogram to be written to the file.
		pathFilename: Location for the file of the waveform output.
		lengthWaveform: n.b. Not optional: the length of the output waveform in samples.
		sampleRate (44100): The sample rate of the output waveform file. Defaults to 44100.
		**parametersSTFT: Keyword-parameters for the inverse Short-Time Fourier Transform, see `stft`.

	Returns:
		None: But see `writeWAV` for additional notes and caveats.
	"""
	if sampleRate is None:
		sampleRate = parametersUniversal['sampleRate']

	makeDirsSafely(pathFilename)
	waveform = stft(spectrogram, inverse=True, lengthWaveform=lengthWaveform, sampleRate=sampleRate, **parametersSTFT)
	writeWAV(pathFilename, waveform, sampleRate)

# TODO inspect this for integration
def waveformSpectrogramWaveform(callableNeedsSpectrogram):
	@functools.wraps(wrapped=callableNeedsSpectrogram)
	def stft_istft(waveform):
		axisTime=-1
		parametersSTFT={} # uh, I think this will be universal or default settings
		arrayTarget = stft(waveform, inverse=False, indexingAxis=None, **parametersSTFT)
		spectrogram = callableNeedsSpectrogram(arrayTarget)
		return stft(spectrogram, inverse=True, indexingAxis=None, lengthWaveform=waveform.shape[axisTime], **parametersSTFT)
	return stft_istft
